"""Unit tests for logistic_regression module.

(C) 2021 Roman Werpachowski.
"""

import unittest

import numpy as np
from sklearn import linear_model

from cppyml import logistic_regression


class TestConjugateGradientLogisticRegression(unittest.TestCase):

    def setUp(self) -> None:
        np.random.seed(33452345)

    def test_against_sklearn(self):
        n = 100
        d = 5

        X = np.random.randn(n, d)
        w = np.random.randn(d)
        z = np.matmul(X, w) + np.random.randn(n) * 0.1
        y = np.sign(z)
        y[y == 0] = 1
        tested = logistic_regression.ConjugateGradientLogisticRegression()
        lam = 0.2
        tol = 1e-15
        tested.set_lam(lam)
        tested.set_weight_absolute_tolerance(tol)
        tested.set_weight_relative_tolerance(tol)
        actual = tested.fit(X, y)
        self.assertTrue(actual.converged)
        benchmark = linear_model.LogisticRegression(tol=tol, C=1/lam)
        expected = benchmark.fit(X, (y + 1) / 2)
        np.testing.assert_allclose(expected.coef_[0], actual.w, rtol=2e-2)

    def test_setters(self):
        lr = logistic_regression.ConjugateGradientLogisticRegression()
        lam = 0.1
        rtol = 1e-8
        atol = 1e-14
        max_steps = 100
        lr.set_lam(lam)
        lr.set_weight_absolute_tolerance(atol)
        lr.set_weight_relative_tolerance(rtol)
        lr.set_maximum_steps(max_steps)
        self.assertEqual(lam, lr.lam)
        self.assertEqual(rtol, lr.weight_relative_tolerance)
        self.assertEqual(atol, lr.weight_absolute_tolerance)
        self.assertEqual(max_steps, lr.maximum_steps)


if __name__ == "__main__":
    unittest.main()



"""Utility functions for the package.

(C) 2021 Roman Werpachowski.
"""
import numpy as np


def standardise_features(features: np.ndarray) -> np.ndarray:
    """Standardises an N x D feature matrix with data points in rows.

    Subtracts a mean feature vector from every row. Divides every column by its
    *biased* standard deviation (if N > 1).

    Args:
        features: N x D feature matrix.

    Returns:
        A standardised copy.
    """
    if len(features.shape) != 2:
        raise ValueError(f"Features matrix must be 2D, got {features.shape}")
    X = features.copy()
    if not X.size:
        return X
    X -= np.mean(X, axis=0)
    if X.shape[0] > 1:
        X /= np.std(X, axis=0, ddof=0)
    return X



"""Unit tests for utils module.

(C) 2020 Roman Werpachowski.
"""
import numpy as np
import unittest

from cppyml import utils


class TestStandardiseFeatures(unittest.TestCase):

    def test_empty(self):
        X = np.array([[]])
        result = utils.standardise_features(X)
        self.assertNotEqual(id(X), id(result))
        np.testing.assert_equal(result, X)

    def test_wrong_shape(self):
        X = np.array([0, 1])
        with self.assertRaises(ValueError):
            utils.standardise_features(X)

    def test_one_row(self):
        X = np.array([[-0.2, 0, 2]])
        expected = np.array([[0, 0, 0]])
        actual = utils.standardise_features(X)
        np.testing.assert_equal(actual, expected)

    def test_two_rows(self):
        X = np.array([
            [-0.2, 0, 2],
            [0.2, 1, 3],
        ])
        mean_row = (X[0] + X[1]) / 2
        stds = np.array([
            np.std(X[:, 0], ddof=0),
            np.std(X[:, 1], ddof=0),
            np.std(X[:, 2], ddof=0),
            ])
        expected = np.array([
            (X[0] - mean_row) / stds,
            (X[1] - mean_row) / stds,
        ])
        actual = utils.standardise_features(X)
        np.testing.assert_almost_equal(actual, expected, decimal=12)


if __name__ == "__main__": 
    unittest.main()



"""Unit tests for clustering module.

(C) 2020-21 Roman Werpachowski.
"""
import unittest
import warnings

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import sklearn.cluster
import sklearn.mixture

from cppyml import clustering


class ClusteringTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.seed = 999
        np.random.seed(cls.seed)
        cls.dims = 2
        sample_size = 1000
        face_radius = 1
        ear_radius = 0.3
        ear_angle = np.pi / 4  # How far is ear from top of head.
        cls.num_components = 3
        radii = (face_radius, ear_radius, ear_radius)
        ear_weight = 2
        weights = [face_radius ** 2, ear_weight * ear_radius ** 2, ear_weight * ear_radius ** 2]
        probabilities = weights / np.sum(weights)

        indices = np.random.choice(np.arange(cls.num_components), sample_size, p=probabilities)

        center_xs = [0, (face_radius + ear_radius) * np.sin(-ear_angle), (face_radius + ear_radius) * np.sin(ear_angle)]
        center_ys = [0, (face_radius + ear_radius) * np.cos(-ear_angle), (face_radius + ear_radius) * np.cos(ear_angle)]

        cls.data = np.empty((sample_size, cls.dims))
        for i in range(sample_size):
            k = indices[i]
            phi = np.random.rand() * 2 * np.pi
            r = np.sqrt(np.random.rand()) * radii[k]
            cls.data[i, 0] = center_xs[k] + r * np.cos(phi)
            cls.data[i, 1] = center_ys[k] + r * np.sin(phi)

    def test_em(self):
        abs_tol = 1e-10
        max_iter = 1000
    
        em = clustering.EM(self.num_components)
        em.set_seed(42)
        em.set_absolute_tolerance(abs_tol)
        em.set_relative_tolerance(0)
        em.set_means_initialiser(clustering.KPP())
        em.set_maximum_steps(max_iter)
        converged = int(em.fit(self.data))
        self.assertTrue(converged)
        pyml_ll = em.log_likelihood

        warnings.filterwarnings("error")
        gmm = sklearn.mixture.GaussianMixture(self.num_components, tol=abs_tol, max_iter=max_iter,
                                              random_state=self.seed, n_init=1, reg_covar=1e-15)
        gmm.fit(self.data)  # gmm raises a warning if not converged.
        sklearn_ll = gmm.score(self.data)

        self.assertAlmostEqual(sklearn_ll, pyml_ll, delta=1e-10)

        u = em.assign_responsibilities(np.array([0, 0]))
        self.assertEqual(3, len(u))
        self.assertAlmostEqual(1, sum(u), delta=1e-15)
        self.assertLessEqual(0, min(u))
        # (0, 0) is the middle of face.
        self.assertAlmostEqual(1, max(u), delta=1e-9)

    def test_k_means(self):
        abs_tol = 1e-10
        max_iter = 1000

        km = clustering.KMeans(self.num_components)
        km.set_seed(42)
        km.set_absolute_tolerance(abs_tol)
        km.set_centroids_initialiser(clustering.KPP())
        km.set_maximum_steps(max_iter)
        km.set_number_initialisations(10)
        converged = int(km.fit(self.data))
        self.assertTrue(converged)
        self.assertGreater(km.inertia, 0)
        self.assertEqual(0, min(km.labels))
        self.assertEqual(self.num_components - 1, max(km.labels))
        self.assertTupleEqual((self.num_components, self.dims), km.centroids.shape)
        for i, centroid in enumerate(km.centroids):
            label, distance = km.assign_label(centroid)
            self.assertEqual(i, label)
            self.assertEqual(0, distance)


if __name__ == "__main__": 
    unittest.main()



"""Unit tests for decision_trees module.

(C) 2020 Roman Werpachowski.
"""

import unittest

import numpy as np
from sklearn.datasets import load_diabetes, load_iris
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

from cppyml import decision_trees


class DecisionTreesTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.diabetes = load_diabetes(return_X_y=True)

    def setUp(self):
        np.random.seed(1066)
        self.min_split_size = 5        

    def test_classification(self):
        iris = load_iris()

        clf = DecisionTreeClassifier(random_state=0, min_samples_split=self.min_split_size)
        clf.fit(iris.data, iris.target)
        sklearn_accuracy = clf.score(iris.data, iris.target)

        tree, _, _ = decision_trees.classification_tree(iris.data, iris.target, alphas=[], min_split_size=self.min_split_size)
        pyml_accuracy = decision_trees.classification_tree_accuracy(tree, iris.data, iris.target)        
        
        self.assertAlmostEqual(sklearn_accuracy, pyml_accuracy, delta=1e-12)

    def test_regression(self):
        X, y = self.diabetes

        regressor = DecisionTreeRegressor(random_state=0, min_samples_split=self.min_split_size)
        regressor.fit(X, y)
        y_hat = regressor.predict(X)
        sklearn_mse = ((y - y_hat)**2).mean()

        tree, _, _ = decision_trees.regression_tree(X, y, alphas=[], min_split_size=self.min_split_size)
        pyml_mse = decision_trees.regression_tree_mean_squared_error(tree, X, y)
        self.assertGreater(pyml_mse, 0)        
        self.assertAlmostEqual(sklearn_mse, pyml_mse, delta=1e-12)

    def test_regression_with_pruning(self):
        X, y = self.diabetes
        tree, _, _ = decision_trees.regression_tree(X, y, alphas=[], min_split_size=self.min_split_size)
        mse = decision_trees.regression_tree_mean_squared_error(tree, X, y)
        pruned_tree, _, _ = decision_trees.regression_tree(X, y, min_split_size=self.min_split_size)
        pruned_mse = decision_trees.regression_tree_mean_squared_error(pruned_tree, X, y)        
        self.assertGreater(pruned_mse, mse)


if __name__ == "__main__": 
    unittest.main()


"""Unit tests for linear_regression module.

(C) 2020 Roman Werpachowski.
"""

import unittest

import numpy as np
from scipy import stats
from sklearn import linear_model

from cppyml import linear_regression
from cppyml import utils


class LinearRegressionTest(unittest.TestCase):

    def setUp(self):
        np.random.seed(4098549032)

    def test_univariate_ols_result(self):
        n = 100
        dof = 98
        rss = 0.01
        tss = 0.1
        slope = -0.1
        intercept = 10
        var_slope = 0.001
        var_intercept = 0.002
        cov_slope_intercept = -0.0005
        var_y = rss / dof
        r2 = 1 - rss / tss
        adjusted_r2 = 1 - (rss / dof) / (tss / (n - 1))
        result = linear_regression.UnivariateOLSResult(
            n, dof, rss, tss, slope, intercept, var_slope, var_intercept, cov_slope_intercept)
        self.assertEqual(n, result.n)
        self.assertEqual(dof, result.dof)
        self.assertEqual(var_y, result.var_y)
        self.assertEqual(r2, result.r2)
        self.assertEqual(adjusted_r2, result.adjusted_r2)
        self.assertEqual(slope, result.slope)
        self.assertEqual(intercept, result.intercept)
        self.assertEqual(var_slope, result.var_slope)
        self.assertEqual(var_intercept, result.var_intercept)
        self.assertEqual(cov_slope_intercept, result.cov_slope_intercept)

    def test_multivariate_ols_result(self):
        n = 100
        dof = 98        
        rss = 0.01
        tss = 0.1
        var_y = rss / dof
        r2 = 1 - rss / tss
        adjusted_r2 = 1 - (rss / dof) / (tss / (n - 1))
        beta = np.array([-0.4, 0.2])
        cov = np.array([[0.004, -0.0001],
                        [-0.00010001, 0.03]])
        result = linear_regression.MultivariateOLSResult(
            n, dof, rss, tss, beta, cov)
        self.assertTrue(result.cov.flags["C_CONTIGUOUS"])
        self.assertEqual(n, result.n)
        self.assertEqual(dof, result.dof)
        self.assertEqual(var_y, result.var_y)
        self.assertEqual(r2, result.r2)
        self.assertEqual(adjusted_r2, result.adjusted_r2)
        np.testing.assert_array_equal(beta, result.beta)
        np.testing.assert_array_equal(cov, result.cov)

    def test_ridge_regression_result(self):
        n = 100
        dof = 98
        rss = 0.01
        tss = 0.1
        var_y = rss / dof
        r2 = 1 - rss / tss
        adjusted_r2 = 1 - (rss / dof) / (tss / (n - 1))
        beta = np.array([-0.4, 0.2])
        cov = np.array([[0.004, -0.0001],
                        [-0.00010001, 0.03]])
        effective_dof = 99
        result = linear_regression.RidgeRegressionResult(
            n, dof, rss, tss, beta, effective_dof, cov)
        self.assertTrue(result.cov.flags["C_CONTIGUOUS"])
        self.assertEqual(n, result.n)
        self.assertEqual(dof, result.dof)
        self.assertEqual(var_y, result.var_y)
        self.assertEqual(r2, result.r2)
        self.assertEqual(adjusted_r2, result.adjusted_r2)
        np.testing.assert_array_equal(beta, result.beta)
        np.testing.assert_array_equal(cov, result.cov)
        self.assertEqual(effective_dof, result.effective_dof)

    def test_lasso_regression_result(self):
        n = 100
        dof = 98
        rss = 0.01
        tss = 0.1
        var_y = rss / dof
        r2 = 1 - rss / tss
        adjusted_r2 = 1 - (rss / dof) / (tss / (n - 1))
        beta = np.array([-0.4, 0.2])
        effective_dof = 99
        result = linear_regression.LassoRegressionResult(
            n, dof, rss, tss, beta, effective_dof)
        self.assertEqual(n, result.n)
        self.assertEqual(dof, result.dof)
        self.assertEqual(var_y, result.var_y)
        self.assertEqual(r2, result.r2)
        self.assertEqual(adjusted_r2, result.adjusted_r2)
        np.testing.assert_array_equal(beta, result.beta)
        self.assertEqual(effective_dof, result.effective_dof)

    def test_univariate_with_intercept(self):
        n = 25
        slope = 0.1
        intercept = -0.9
        noise = 0.1
        x = np.random.randn(n)
        y = x * slope + intercept + noise * np.random.randn(n)
        result = linear_regression.univariate(x, y)
        slope, intercept, rvalue, _, stderr = stats.linregress(x, y)
        self.assertEqual(n, result.n)
        self.assertEqual(n - 2, result.dof)
        self.assertAlmostEqual(slope, result.slope, delta=1e-15)
        self.assertAlmostEqual(intercept, result.intercept, delta=1e-15)
        self.assertAlmostEqual(rvalue**2, result.r2, delta=1e-15)
        adjusted_r2 = 1 - (n - 1) * (1 - rvalue**2) / (n - 2)
        self.assertAlmostEqual(adjusted_r2, result.adjusted_r2, delta=1e-15)
        self.assertAlmostEqual(stderr**2, result.var_slope, delta=1e-15)

    def test_univariate_with_intercept_regular(self):
        n = 25
        slope = 0.1
        intercept = -0.9
        noise = 0.1
        x0 = - 0.4
        dx = 0.04
        x = np.linspace(x0, x0 + dx * (n - 1), n)
        y = x * slope + intercept + noise * np.random.randn(n)
        result = linear_regression.univariate_regular(x0, dx, y)
        slope, intercept, rvalue, _, stderr = stats.linregress(x, y)
        self.assertEqual(n, result.n)
        self.assertEqual(n - 2, result.dof)
        self.assertAlmostEqual(slope, result.slope, delta=1e-15)
        self.assertAlmostEqual(intercept, result.intercept, delta=1e-15)
        self.assertAlmostEqual(rvalue**2, result.r2, delta=1e-15)
        adjusted_r2 = 1 - (n - 1) * (1 - rvalue**2) / (n - 2)
        self.assertAlmostEqual(adjusted_r2, result.adjusted_r2, delta=1e-15)
        self.assertAlmostEqual(stderr**2, result.var_slope, delta=1e-15)

    def test_univariate_without_intercept(self):
        n = 25
        slope = 0.1
        noise = 0.1
        x = np.random.randn(n)
        y = x * slope + noise * np.random.randn(n)
        result = linear_regression.univariate_without_intercept(x, y)
        X = np.empty((n, 1), dtype=float)
        X[:, 0] = x
        lr = linear_model.LinearRegression(fit_intercept=False)
        lr.fit(X, y)
        self.assertEqual(n, result.n)
        self.assertEqual(n - 1, result.dof)
        self.assertAlmostEqual(lr.coef_[0], result.slope, delta=1e-15)
        self.assertEqual(0, result.intercept)
        self.assertEqual(0, result.var_intercept)
        self.assertEqual(0, result.cov_slope_intercept)
        self.assertAlmostEqual(result.r2, result.adjusted_r2, delta=1e-15)

    def test_multivariate_add_ones(self):
        n = 25
        d = 4
        X = np.random.randn(n, d)
        y = -0.5 * X[:, 0] + 0.1 * X[:, 1] - X[:, 3] + 4 + 0.2 * np.random.randn(n)        
        result = linear_regression.multivariate(X, y, True)
        self.assertEqual(n, result.n)
        self.assertEqual(n - d - 1, result.dof)
        lr = linear_model.LinearRegression()
        lr.fit(X, y)
        np.testing.assert_array_almost_equal(result.beta[:-1], lr.coef_, decimal=15)
        self.assertAlmostEqual(result.beta[-1], lr.intercept_, delta=1e-15)
        adjusted_r2 = 1 - (n - 1) * (1 - result.r2) / (n - d - 1)
        self.assertAlmostEqual(adjusted_r2, result.adjusted_r2, delta=1e-15)

    def test_multivariate(self):
        n = 25
        d = 5
        X = np.random.randn(n, d)
        X[:, d - 1] = 1
        y = -0.5 * X[:, 0] + 0.1 * X[:, 1] - X[:, 3] + 4 + 0.2 * np.random.randn(n)
        result = linear_regression.multivariate(X, y, False)
        self.assertTrue(result.cov.flags["C_CONTIGUOUS"])
        self.assertEqual(n, result.n)
        self.assertEqual(n - d, result.dof)
        lr = linear_model.LinearRegression(fit_intercept=False)  # Force lr not to add another column with 1's.
        lr.fit(X, y)
        np.testing.assert_array_almost_equal(result.beta, lr.coef_, decimal=14)
        adjusted_r2 = 1 - (n - 1) * (1 - result.r2) / (n - d)
        self.assertAlmostEqual(adjusted_r2, result.adjusted_r2, delta=1e-15)

    def test_recursive_multivariate_ols_no_data(self):
        rmols = linear_regression.RecursiveMultivariateOLS()
        self.assertEqual(0, rmols.n)
        self.assertEqual(0, rmols.d)
        self.assertEqual(0, len(rmols.beta))

    def test_recursive_multivariate_ols_errors_init(self):
        with self.assertRaises(ValueError):
            linear_regression.RecursiveMultivariateOLS(np.zeros((5, 10)), np.zeros(5))
        with self.assertRaises(ValueError):
            linear_regression.RecursiveMultivariateOLS(np.zeros((10, 10)), np.zeros(12))

    def test_recursive_multivariate_ols_errors_update(self):
        rmols = linear_regression.RecursiveMultivariateOLS()
        with self.assertRaises(ValueError):
            rmols.update(np.zeros((5, 10)), np.zeros(5))
        with self.assertRaises(ValueError):
            rmols.update(np.zeros((10, 10)), np.zeros(12))

    def test_recursive_multivariate_ols_errors_one_sample(self):
        n = 10
        d = 3
        X = np.random.randn(n, d)
        true_beta = 0.5 - np.random.rand(d)
        y = np.matmul(X, true_beta) + 0.1 * np.random.randn(n)
        rmols1 = linear_regression.RecursiveMultivariateOLS()
        rmols1.update(X, y)
        self.assertEqual(n, rmols1.n)
        self.assertEqual(d, rmols1.d)
        rmols2 = linear_regression.RecursiveMultivariateOLS(X, y)
        self.assertEqual(n, rmols2.n)
        self.assertEqual(d, rmols2.d)
        np.testing.assert_array_equal(rmols1.beta, rmols2.beta)
        expected_beta = linear_regression.multivariate(X, y).beta
        np.testing.assert_array_almost_equal(expected_beta, rmols1.beta, 16)

    def test_recursive_multivariate_ols_many_samples(self):
        d = 10
        true_beta = 0.5 - np.random.rand(d)
        sample_sizes = [d, 4, 20, 6, 20, 4, 1, 100]
        rmols = linear_regression.RecursiveMultivariateOLS()
        cumulative_n = 0
        total_n = sum(sample_sizes)
        all_X = np.random.randn(total_n, d)
        all_y = np.matmul(all_X, true_beta) + 0.1 * np.random.randn(total_n)
        for sample_idx, n in enumerate(sample_sizes):
            X = all_X[cumulative_n : (cumulative_n + n)]
            y = all_y[cumulative_n : (cumulative_n + n)]
            rmols.update(X, y)
            cumulative_n += n
            cumulative_X = all_X[:cumulative_n]
            cumulative_y = all_y[:cumulative_n]
            ols_beta = linear_regression.multivariate(cumulative_X, cumulative_y).beta
            np.testing.assert_array_almost_equal(ols_beta, rmols.beta, 13)

    def test_ridge(self):
        n =25
        d = 4
        X = np.random.randn(n, d)
        true_beta = np.random.rand(d)
        intercept = 0.2
        y = np.matmul(X, true_beta) + 0.4 * np.random.randn(n) + intercept
        means = np.mean(X, axis=0)
        Xstd = X - means
        stds = np.std(Xstd, axis=0, ddof=0)
        Xstd = Xstd / stds
        lam = 0.01
        ridge = linear_model.Ridge(alpha=lam, fit_intercept=True, normalize=False)
        result1 = linear_regression.ridge(Xstd, y, lam, do_standardise=False)
        ridge.fit(Xstd, y)
        result2 = linear_regression.ridge(Xstd, y, lam, do_standardise=True)
        result3 = linear_regression.ridge(X, y, lam, do_standardise=True)
        for result in (result1, result2, result3):
            self.assertEqual(n, result.n)
            self.assertEqual(n - d - 1, result.dof)            
        for idx, result in enumerate((result1, result2)):
            self.assertEqual(d + 1, len(result.beta), msg=str(idx))
            np.testing.assert_array_almost_equal(ridge.coef_, result.beta[:d], 14, err_msg=str(idx))            
        np.testing.assert_array_almost_equal(ridge.coef_ / stds, result3.beta[:d], 14)
        self.assertAlmostEqual(ridge.intercept_, result1.beta[d], delta=1e-14)
        self.assertAlmostEqual(ridge.intercept_, result2.beta[d], delta=1e-14)
        self.assertAlmostEqual(ridge.intercept_ - np.dot(ridge.coef_, means / stds), result3.beta[d], delta=1e-14)
        r2 = ridge.score(Xstd, y)
        self.assertAlmostEqual(r2, result1.r2, delta=1e-14)
        self.assertAlmostEqual(r2, result2.r2, delta=1e-14)
        self.assertAlmostEqual(r2, result3.r2, delta=1e-14)
        adjusted_r2 = 1 - (1 - r2) * (n - 1) / (n - d - 1)
        self.assertAlmostEqual(adjusted_r2, result1.adjusted_r2, delta=1e-14)
        self.assertAlmostEqual(adjusted_r2, result2.adjusted_r2, delta=1e-14)
        self.assertAlmostEqual(adjusted_r2, result3.adjusted_r2, delta=1e-14)

    def test_lasso_without_standardisation_polynomial(self):
        n = 101
        d = 3
        X = np.empty((n, d), dtype=float)
        x = np.linspace(-1, 1, n)
        for i in range(d):
            X[:, i] = x ** (i + 1)
        y = np.abs(x) + 1 + 0.2 * np.sin(x)
        X = utils.standardise_features(X)
        lam = 0.1
        result = linear_regression.lasso(X, y, lam * 2 * n, do_standardise=False)
        lasso = linear_model.Lasso(alpha=lam, fit_intercept=True, normalize=False)
        lasso.fit(X, y)
        np.testing.assert_almost_equal(result.beta[:d], lasso.coef_, decimal=13)
        self.assertAlmostEqual(lasso.intercept_, result.beta[d], delta=1e-13)
        r2 = lasso.score(X, y)
        self.assertAlmostEqual(r2, result.r2, delta=1e-13)

    def test_press_univariate_with_intercept(self):
        x = np.array([-1, 0, 1])
        y = np.array([1, 0, 1])
        actual1 = linear_regression.press_univariate(x, y, True)
        actual2 = linear_regression.press_univariate(x, y)
        self.assertEqual(actual1, actual2)
        self.assertAlmostEqual(9, actual1, delta=1e-15)

    def test_press_univariate_without_intercept(self):
        x = np.array([-1, 0, 1])
        y = np.array([1, 0, 1])
        actual = linear_regression.press_univariate(x, y, False)
        self.assertAlmostEqual(8, actual, delta=1e-15)

    def test_press_no_regularisation(self):
        X = np.array([[-1, 1], [0, 1], [1, 1]])
        y = np.array([1, 0, 1])
        actual1 = linear_regression.press(X, y)
        actual2 = linear_regression.press(X, y, "none")        
        self.assertEqual(actual1, actual2)
        self.assertAlmostEqual(9, actual1, delta=1e-15)

    def test_press_ridge_zero_strength(self):
        X = np.array([[-1], [0], [1]])
        y = np.array([1, 0, 1])
        actual1 = linear_regression.press(X, y, "ridge")
        actual2 = linear_regression.press(X, y, "ridge", 0)
        self.assertEqual(actual1, actual2)
        self.assertAlmostEqual(9, actual1, delta=1e-15)

    def test_press_ridge(self):
        X = np.array([[-1], [0], [1]])
        y = np.array([1, 0, 1])
        actual = linear_regression.press(X, y, "ridge", 0.1)
        self.assertGreater(9, actual)


if __name__ == "__main__": 
    unittest.main()



"""Setup script for cppyml package.

(C) 2020 Roman Werpachowski.

To install the package, run "python setup.py install" from cppyml
directory.

To build a distributable version, run "python setup.py sdist bdist_wheel".
"""
from setuptools.dist import Distribution
from setuptools import setup, find_packages
import argparse
import json
import glob
import os
import sys
import shutil

argparser = argparse.ArgumentParser(add_help=False)
argparser.add_argument("--debug", dest="debug", default=False, action="store_true",
                       help="Use C++ debug build")
args, unknown = argparser.parse_known_args()
sys.argv = [sys.argv[0]] + unknown

# Whether to install a Debug version of the binary files.
DEBUG_BINARIES = args.debug

# Name of the C++ build mode.
CPP_BUILD_MODE = "Debug" if DEBUG_BINARIES else "Release"

# Package version.
with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "version.json")) as f:
    VERSION = json.load(f)

# Package name.
NAME = "cppyml"

# Directory where setup.py is located.
SETUP_DIRNAME = os.path.abspath(os.path.dirname(__file__))

# Base directory of the whole project.
BASE_DIRECTORY = os.path.abspath(os.path.join(SETUP_DIRNAME, ".."))

# Architecture is 64-bit by default.

# Filenames and paths of binary files needed.
if os.name == "posix":
    PYML_FILENAME = "cppyml.so"   
    PYML_PATH = os.path.join(BASE_DIRECTORY, "cppyml",
                             "build", CPP_BUILD_MODE, PYML_FILENAME)
else:
    PYML_FILENAME = "cppyml.pyd"
    BINARY_DIRECTORY = os.path.join(BASE_DIRECTORY, "x64", CPP_BUILD_MODE + "Static")
    PYML_PATH = os.path.join(BINARY_DIRECTORY, PYML_FILENAME)

    
BINARY_FILENAMES = [PYML_FILENAME]
SRC_PATHS = [PYML_PATH]


def get_binary_files():
    package_dirname = os.path.join(SETUP_DIRNAME, NAME)
    # Remove old files.
    for extension in ["dll", "pyd", "so"]:
        paths = glob.glob(os.path.join(package_dirname, "*.%s" % extension))
        for path in paths:
            try:
                os.remove(path)
            except FileNotFoundError:
                print("File %s does not exist", path)
            else:
                print("Removed %s" % path)
    for src_path, dst_filename in zip(SRC_PATHS, BINARY_FILENAMES):
        dst_path = os.path.join(package_dirname, dst_filename)
        shutil.copyfile(src_path, dst_path)
        print("Copied %s to %s" % (src_path, dst_path))


def load_readme():
    with open(os.path.join(SETUP_DIRNAME, "README.md"), "r") as f:
        return f.read()


get_binary_files()

PACKAGE_DATA = BINARY_FILENAMES


class BinaryDistribution(Distribution):
    """Distribution which always forces a binary package with platform name"""
    def has_ext_modules(foo):
        return True


setup(
    name=NAME,
    author="Roman Werpachowski",
    url="https://romanwerpachowski.github.io/ML/",
    author_email="roman.werpachowski@gmail.com",
    license="GPL-3.0",
    keywords="machine-learning ML extension algorithms numerical optimised",
    version=VERSION,
    description="Efficient implementations of selected ML algorithms for Python.",
    long_description=load_readme(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    package_data={NAME: PACKAGE_DATA},
    distclass=BinaryDistribution,
)
